#ifndef POINT2D_H
#define POINT2D_H

template<typename T> class CPoint2D
{
public:
	T x;
	T y;

	CPoint2D() : x(0), y(0) {}
	CPoint2D(T aX, T aY) : x(aX), y(aY) {}
	CPoint2D(const CPoint2D& pt) : x(pt.x), y(pt.y) {}
	virtual ~CPoint2D() {}
	CPoint2D& operator =(const CPoint2D& pt){ x = pt.x; y = pt.y; return *this; }
	CPoint2D& operator +=(const CPoint2D& pt){ x += pt.x; y += pt.y; return *this; }
	CPoint2D& operator -=(const CPoint2D& pt){ x -= pt.x; y -= pt.y; return *this; }
	CPoint2D& operator *=(T val){ x *= val; y *= val; return *this;}
	CPoint2D& operator /=(T val){ if(val != 0){ x /= val; y /= val; } return *this; }

	void LRotate()
	{
		T tmpX = x;
		x = y;
		y = -tmpX;
	}

	void RRotate()
	{
		T tmpX = x;
		x = -y;
		y = tmpX;
	}

	// (1/2, 1/2)𒆐SƂĉ]
	void LRotateC()
	{
		LRotate();
		y += 1;
	}

	void RRotateC()
	{
		RRotate();
		x += 1;
	}

	friend CPoint2D operator +(const CPoint2D& pt1, const CPoint2D& pt2)
	{
		CPoint2D pt(pt1);
		pt += pt2;
		return pt;
	}

	friend CPoint2D operator -(const CPoint2D& pt1, const CPoint2D& pt2)
	{
		CPoint2D pt(pt1);
		pt -= pt2;
		return pt;
	}

	friend CPoint2D operator *(const CPoint2D& pt1, T val)
	{
		CPoint2D pt(pt1);
		pt *= val;
		return pt;
	}

	friend CPoint2D operator /(const CPoint2D& pt1, T val)
	{
		CPoint2D pt(pt1);
		pt /= val;
		return pt;
	}

	friend bool operator ==(const CPoint2D& pt1, const CPoint2D& pt2)
	{
		if((pt1.x == pt2.x) && (pt1.y == pt2.y)){
			return true;
		}else{
			return false;
		}
	}

	friend bool operator !=(const CPoint2D& pt1, const CPoint2D& pt2)
	{
		return !(pt1 == pt2);
	}

	static T InnerProduct(const CPoint2D& pt1, const CPoint2D& pt2)
	{
		return (T)(pt1.x * pt2.x + pt1.y + pt2.y);
	}
};

#endif // POINT2D_H
