#include "SRSMino.h"
#include <assert.h>

// e[uLqȗ̂߁Atypedef
typedef CPoint2D<int> CPt;

// For IMino
const CSRSMino::SSRSWallKickTable CSRSMino::s_tableIMinoL = {
	{4, CPt(2, 0), CPt(-1, 0), CPt(-1, -2), CPt(2, 1)}, // from0
	{4, CPt(2, 0), CPt(-1, 0), CPt(2, -1), CPt(-1, 2)}, // from90
	{4, CPt(-2, 0), CPt(1, 0), CPt(-2, -1), CPt(1, 1)}, // from180
	{4, CPt(1, 0), CPt(-2, 0), CPt(1, -2), CPt(-2, 1)}  // from270
};

const CSRSMino::SSRSWallKickTable CSRSMino::s_tableIMinoR = {
	{4, CPt(-2, 0), CPt(1, 0), CPt(1, -2), CPt(-2, 1)}, // from0
	{4, CPt(-1, 0), CPt(2, 0), CPt(-1, -2), CPt(2, 1)}, // from90
	{4, CPt(2, 0), CPt(-1, 0), CPt(2, -1), CPt(-1, 1)}, // from180
	{4, CPt(-2, 0), CPt(1, 0), CPt(-2, -1), CPt(1, 2)}  // from270
};

// For J/L/S/Z Mino
const CSRSMino::SSRSWallKickTable CSRSMino::s_tableCommonL = {
	{4, CPt(1, 0), CPt(1, -1), CPt(0, 2), CPt(1, 2)},     // from0
	{4, CPt(1, 0), CPt(1, 1), CPt(0, -2), CPt(1, -2)},    // from90
	{4, CPt(-1, 0), CPt(-1, -1), CPt(0, 2), CPt(-1, 2)},  // from180
	{4, CPt(-1, 0), CPt(-1, 1), CPt(0, -2), CPt(-1, -2)}  // from270
};

const CSRSMino::SSRSWallKickTable CSRSMino::s_tableCommonR = {
	{4, CPt(-1, 0), CPt(-1, -1), CPt(0, 2), CPt(-1, 2)},  // from0
	{4, CPt(1, 0), CPt(1, 1), CPt(0, -2), CPt(1, -2)},    // from90
	{4, CPt(1, 0), CPt(1, -1), CPt(0, 2), CPt(1, 2)},     // from180
	{4, CPt(-1, 0), CPt(-1, 1), CPt(0, -2), CPt(-1, -2)}  // from270
};

// For T Mino
const CSRSMino::SSRSWallKickTable CSRSMino::s_tableTMinoL = {
	{3, CPt(1, 0), CPt(1, -1), /*CPt(0, 2),*/ CPt(1, 2)},    // from0
	{4, CPt(1, 0), CPt(1, 1), CPt(0, -2), CPt(1, -2)},       // from90
	{3, CPt(-1, 0), /*CPt(-1, -1),*/ CPt(0, 2), CPt(-1, 2)}, // from180
	{4, CPt(-1, 0), CPt(-1, 1), CPt(0, -2), CPt(-1, -2)}     // from270
};

const CSRSMino::SSRSWallKickTable CSRSMino::s_tableTMinoR = {
	{3, CPt(-1, 0), CPt(-1, -1), /*CPt(0, 2),*/ CPt(-1, 2)}, // from0
	{4, CPt(1, 0), CPt(1, 1), CPt(0, -2), CPt(1, -2)},       // from90
	{3, CPt(1, 0), /*CPt(1, -1),*/ CPt(0, 2), CPt(1, 2)},    // from180
	{4, CPt(-1, 0), CPt(-1, 1), CPt(0, -2), CPt(-1, -2)}     // from270
};

CSRSMino::CSRSMino()
{
	m_pMinoBlocks = new CMinoBlocks(4);
	m_pTableL = m_pTableR = NULL; // hNXœK؂ɐݒ
}

CSRSMino::~CSRSMino()
{
	delete m_pMinoBlocks;
}

CMino::MinoResult CSRSMino::LRotate()
{
	if(!m_pField || !m_pMinoBlocks || !m_pTableL){
		return 0; // NULLł͂Ȃ
	}

	// ݏ󋵂ۑĂ
	ERotateState oldRotateState = m_rotateState;
	CMinoBlocks oldMinoBlocks(*m_pMinoBlocks);
	ERotateState newRotateState = GetLRotatedState(m_rotateState);

	LRotateNormal(); // ܂ʂɉ]

	// ܂̂܂܂Ń`FbNI
	if(m_pField->CheckOverlap(*m_pMinoBlocks) == CTetrisField::OK){
		m_rotateState = newRotateState;
		m_bWallKick = false;
		UpdateGhostMinoBlocks();
		return CMino::SUCCESS;
	}
	// ɂ炵ă`FbNI
	{
		const SSRSWallKickOrder *pKickOrder = NULL;
		switch(oldRotateState){
		case ROTATE0:
		default:
			pKickOrder = &m_pTableL->from0; break;
		case ROTATE90:
			pKickOrder = &m_pTableL->from90; break;
		case ROTATE180:
			pKickOrder = &m_pTableL->from180; break;
		case ROTATE270:
			pKickOrder = &m_pTableL->from270; break;
		}

		unsigned int i = 0;
		for(i=0; i<pKickOrder->num; i++){
			m_pMinoBlocks->GetBaseCoordinate() += pKickOrder->move[i];
			if(m_pField->CheckOverlap(*m_pMinoBlocks) == CTetrisField::OK){
				m_rotateState = newRotateState;
				m_bWallKick = true;
				UpdateGhostMinoBlocks();
				return CMino::SUCCESS;
			}
			// ʖڂ̂Ō̈ʒuɖ߂
			m_pMinoBlocks->GetBaseCoordinate() -= pKickOrder->move[i];
		}
	}

	// ܂Ȃ̂ŁAɖ߂
	m_rotateState = oldRotateState;
	m_pMinoBlocks->Copy(oldMinoBlocks);
	return 0;
}

CMino::MinoResult CSRSMino::RRotate()
{
	if(!m_pField || !m_pMinoBlocks || !m_pTableR){
		return 0; // NULLł͂Ȃ
	}

	// ݏ󋵂ۑĂ
	ERotateState oldRotateState = m_rotateState;
	CMinoBlocks oldMinoBlocks(*m_pMinoBlocks);
	ERotateState newRotateState = GetRRotatedState(m_rotateState);

	RRotateNormal(); // ܂ʂɉ]

	// ܂̂܂܂Ń`FbNI
	if(m_pField->CheckOverlap(*m_pMinoBlocks) == CTetrisField::OK){
		m_rotateState = newRotateState;
		m_bWallKick = false;
		UpdateGhostMinoBlocks();
		return CMino::SUCCESS;
	}
	// ɂ炵ă`FbNI
	{
		const SSRSWallKickOrder *pKickOrder = NULL;
		switch(oldRotateState){
		case ROTATE0:
		default:
			pKickOrder = &m_pTableR->from0; break;
		case ROTATE90:
			pKickOrder = &m_pTableR->from90; break;
		case ROTATE180:
			pKickOrder = &m_pTableR->from180; break;
		case ROTATE270:
			pKickOrder = &m_pTableR->from270; break;
		}

		unsigned int i = 0;
		for(i=0; i<pKickOrder->num; i++){
			m_pMinoBlocks->GetBaseCoordinate() += pKickOrder->move[i];
			if(m_pField->CheckOverlap(*m_pMinoBlocks) == CTetrisField::OK){
				m_rotateState = newRotateState;
				m_bWallKick = true;
				UpdateGhostMinoBlocks();
				return CMino::SUCCESS;
			}
			// ʖڂ̂Ō̈ʒuɖ߂
			m_pMinoBlocks->GetBaseCoordinate() -= pKickOrder->move[i];
		}
	}

	// ܂Ȃ̂ŁAɖ߂
	m_rotateState = oldRotateState;
	m_pMinoBlocks->Copy(oldMinoBlocks);
	return 0;
}

// IMino
// 
CMino::MinoResult CSRSIMino::Init()
{
	assert(m_pMinoBlocks != NULL);
	assert(m_pMinoBlocks->GetBlockNum() == 4);

	m_pMinoBlocks->GetBaseCoordinate() = CPoint2D<int>(0, 0);

	m_rotateState = ROTATE0;
	m_pMinoBlocks->SetBlock(0, CPoint2D<int>(0, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(1, CPoint2D<int>(1, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(2, CPoint2D<int>(2, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(3, CPoint2D<int>(-1, 0), s_blockState, false);

	return CMino::SUCCESS;
}

// SMino
/*
@@
@
*/
CMino::MinoResult CSRSSMino::Init()
{
	assert(m_pMinoBlocks != NULL);
	assert(m_pMinoBlocks->GetBlockNum() == 4);

	m_pMinoBlocks->GetBaseCoordinate() = CPoint2D<int>(0, 0);

	m_rotateState = ROTATE0;
	m_pMinoBlocks->SetBlock(0, CPoint2D<int>(0, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(1, CPoint2D<int>(-1, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(2, CPoint2D<int>(0, -1), s_blockState, false);
	m_pMinoBlocks->SetBlock(3, CPoint2D<int>(1, -1), s_blockState, false);

	return CMino::SUCCESS;
}

// ZMino
/*
@
@@
*/
CMino::MinoResult CSRSZMino::Init()
{
	assert(m_pMinoBlocks != NULL);
	assert(m_pMinoBlocks->GetBlockNum() == 4);

	m_pMinoBlocks->GetBaseCoordinate() = CPoint2D<int>(0, 0);

	m_rotateState = ROTATE0;
	m_pMinoBlocks->SetBlock(0, CPoint2D<int>(0, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(1, CPoint2D<int>(1, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(2, CPoint2D<int>(0, -1), s_blockState, false);
	m_pMinoBlocks->SetBlock(3, CPoint2D<int>(-1, -1), s_blockState, false);

	return CMino::SUCCESS;
}

// JMino
/*
@
@
*/
CMino::MinoResult CSRSJMino::Init()
{
	assert(m_pMinoBlocks != NULL);
	assert(m_pMinoBlocks->GetBlockNum() == 4);

	m_pMinoBlocks->GetBaseCoordinate() = CPoint2D<int>(0, 0);

	m_rotateState = ROTATE0;
	m_pMinoBlocks->SetBlock(0, CPoint2D<int>(0, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(1, CPoint2D<int>(1, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(2, CPoint2D<int>(-1, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(3, CPoint2D<int>(-1, -1), s_blockState, false);

	return CMino::SUCCESS;
}

// LMino
/*
@@@
@
*/
CMino::MinoResult CSRSLMino::Init()
{
	assert(m_pMinoBlocks != NULL);
	assert(m_pMinoBlocks->GetBlockNum() == 4);

	m_pMinoBlocks->GetBaseCoordinate() = CPoint2D<int>(0, 0);

	m_rotateState = ROTATE0;
	m_pMinoBlocks->SetBlock(0, CPoint2D<int>(0, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(1, CPoint2D<int>(1, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(2, CPoint2D<int>(-1, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(3, CPoint2D<int>(1, -1), s_blockState, false);

	return CMino::SUCCESS;
}

// TMino
/*
@@
@
*/
CMino::MinoResult CSRSTMino::Init()
{
	assert(m_pMinoBlocks != NULL);
	assert(m_pMinoBlocks->GetBlockNum() == 4);

	m_pMinoBlocks->GetBaseCoordinate() = CPoint2D<int>(0, 0);

	m_rotateState = ROTATE0;
	m_pMinoBlocks->SetBlock(0, CPoint2D<int>(0, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(1, CPoint2D<int>(1, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(2, CPoint2D<int>(-1, 0), s_blockState, false);
	m_pMinoBlocks->SetBlock(3, CPoint2D<int>(0, -1), s_blockState, false);

	return CMino::SUCCESS;
}

CMino::MinoResult CSRSTMino::LRotate()
{
	MinoResult result = CSRSMino::LRotate();
	if(result & CMino::SUCCESS){
		m_bTSpin = CheckTSpinCondition(m_pMinoBlocks->GetBaseCoordinate());
	}
	return result;
}

CMino::MinoResult CSRSTMino::RRotate()
{
	MinoResult result = CSRSMino::RRotate();
	if(result & CMino::SUCCESS){
		m_bTSpin = CheckTSpinCondition(m_pMinoBlocks->GetBaseCoordinate());
	}
	return result;
}
