#include "TetrisField.h"
#include <assert.h>

CTetrisField::CTetrisField(unsigned int nXLen, unsigned int nYLen, unsigned int nYMarginLen)
{
	m_nXLen = nXLen;
	m_nYLen = nYLen + nYMarginLen;

	m_ppFieldLine = new CTetrisFieldLine*[m_nYLen];
	assert(m_ppFieldLine != NULL);
	
	{
		unsigned int i = 0;
		for(i=0; i<m_nYLen; i++){
			m_ppFieldLine[i] = new CTetrisFieldLine(m_nXLen);
			assert(m_ppFieldLine[i] != NULL);
		}
	}

	m_ptBase.x = 0;
	m_ptBase.y = nYMarginLen;
}

CTetrisField::~CTetrisField()
{
	unsigned int i = 0;
	for(i=0; i<m_nYLen; i++){
		delete m_ppFieldLine[i];
	}

	delete[] m_ppFieldLine;
}

void CTetrisField::Init()
{
	unsigned int i = 0;
	for(i=0; i<m_nYLen; i++){
		m_ppFieldLine[i]->SetBlockState(BLOCKSTATE_NONE);
	}
}

CTetrisField::TetrisFieldResult
CTetrisField::GetBlockState(const CPoint2D<int> &rPt, ETetrisBlockState &rBlockState) const
{
	CPoint2D<int> pt = rPt + m_ptBase; // z̍W
	TetrisFieldResult result = OK;

	rBlockState = BLOCKSTATE_NONE;

	if(pt.x < 0){
		result |= LWALL_OVERLAP;
	}else if(pt.x >= (int)m_nXLen){
		result |= RWALL_OVERLAP;
	}

	if(pt.y < 0){
		result |= CEIL_OVERLAP;
	}else if(pt.y >= (int)m_nYLen){
		result |= FLOOR_OVERLAP;
	}

	if(result == OK){
		rBlockState = m_ppFieldLine[pt.y]->GetBlockState(pt.x);	
	}

	return result;
}

CTetrisField::TetrisFieldResult
CTetrisField::SetBlockState(const CPoint2D<int> &rPt, ETetrisBlockState blockState)
{
	CPoint2D<int> pt = rPt + m_ptBase; // z̍W
	TetrisFieldResult result = OK;

	if(pt.x < 0){
		result |= LWALL_OVERLAP;
	}else if(pt.x >= (int)m_nXLen){
		result |= RWALL_OVERLAP;
	}

	if(pt.y < 0){
		result |= CEIL_OVERLAP;
	}else if(pt.y >= (int)m_nYLen){
		result |= FLOOR_OVERLAP;
	}

	if(result == OK){
		m_ppFieldLine[pt.y]->SetBlockState(pt.x, blockState);	
	}

	return result;
}

CTetrisField::TetrisFieldResult
CTetrisField::CheckOverlap(const CMinoBlocks &rMinoBlocks) const
{
	TetrisFieldResult result = OK;
	CPoint2D<int> pt;
	ETetrisBlockState blockState = BLOCKSTATE_NONE;

	unsigned int i = 0;
	for(i=0; i<rMinoBlocks.GetBlockNum(); i++){
		// ܂MinoBlocksIuWFNgBlock̍W΍WŎ擾
		rMinoBlocks.GetBlock(i, pt, blockState, true);
		// ̍WBlock邩`FbN
/*
		result |= GetBlockState(pt, blockState);
		if((result == OK) && (blockState != BLOCKSTATE_NONE)){
			// ̍WɊBlockL
			result |= BLOCK_OVERLAP;
		}
*/
		result |= CheckOverlap(pt);
	}

	return result;
}

CTetrisField::TetrisFieldResult
CTetrisField::CheckOverlap(const CPoint2D<int> &rPt) const
{
	ETetrisBlockState blockState = BLOCKSTATE_NONE;

	TetrisFieldResult result = GetBlockState(rPt, blockState);
	if((result == OK) && (blockState != BLOCKSTATE_NONE)){
		result = BLOCK_OVERLAP;
	}

	return result;
}

CTetrisField::TetrisFieldResult
CTetrisField::SetBlockState(const CMinoBlocks &rMinoBlocks)
{
	TetrisFieldResult result = OK;
	CPoint2D<int> pt;
	ETetrisBlockState blockState = BLOCKSTATE_NONE;

	unsigned int i = 0;
	for(i=0; i<rMinoBlocks.GetBlockNum(); i++){
		// ܂MinoBlocksIuWFNgBlock̍W΍WŎ擾
		rMinoBlocks.GetBlock(i, pt, blockState, true);
		// ̍WBlocku
		result |= SetBlockState(pt, blockState);
	}

	return result;
}

bool CTetrisField::SetMinoBlocksToStartPosition(CMinoBlocks &rMinoBlocks) const
{
	CPoint2D<int> ptTopLeft;
	CPoint2D<int> ptBottomRight;

	// ܂Oڎlp`߂
	rMinoBlocks.GetCircumscribedRect(ptTopLeft, ptBottomRight, true);

	/*
	@@
	@ <- ptBottomRight.y
	@@
	@@(ptTopLeft.x + ptBottomRight.x)/2
	AW(m_nXLen/2 - 1, -2)ɗ悤Ɉړ
	*/
	CPoint2D<int> vec = CPoint2D<int>(m_nXLen/2 - 1, -2) - CPoint2D<int>((ptTopLeft.x + ptBottomRight.x)/2, ptBottomRight.y);
	rMinoBlocks.GetBaseCoordinate() += vec;

	if(CheckOverlap(rMinoBlocks) == CTetrisField::OK){
		return true;
	}else{
		return false; // Mino傫(oO) or Q[I[o[
	}
}

bool CTetrisField::IsAllClear() const
{
	unsigned int i = 0;
	for(i=0; i<m_nYLen; i++){
		// łłȂCfalse
		if(!m_ppFieldLine[i]->IsEmpty()){
			return false;
		}
	}
	return true;
}

unsigned int CTetrisField::ClearFilledLine()
{
	unsigned int nFilledLineNum = 0;
	unsigned int i = 0;
	for(i=0; i<m_nYLen; i++){
		if(m_ppFieldLine[i]->IsFilled()){
			// ̃CSBlockŖ܂Ă!
			nFilledLineNum++;
			ClearALine(i);
		}
	}

	return nFilledLineNum;
}

// protected
void CTetrisField::ClearALine(unsigned int nLine)
{
	// nLineԖڂ̃CāȀ̃CɗƂ
	if(nLine >= m_nYLen){
		return;
	}

	CTetrisFieldLine *pLineCleared = m_ppFieldLine[nLine];
	pLineCleared->SetBlockState(BLOCKSTATE_NONE);
	unsigned int i = 0;
	for(i=nLine; i>=1; i--){
		// |C^փRs[Ă
		m_ppFieldLine[i] = m_ppFieldLine[i-1];
	}
	m_ppFieldLine[0] = pLineCleared; // ꂽC͈ԏ
}
