#ifndef TETRISFIELD_H
#define TETRISFIELD_H

#include "TetrisCommon.h"
#include "Point2D.h"
#include "MinoBlocks.h"

/*
  m_ppFieldLine[]
  [0]     -------->  CTetrisFieldLineCX^X
  [1]     -------->  CTetrisFieldLineCX^X
  [2]     -------->  CTetrisFieldLineCX^X
  [3]     -------->  CTetrisFieldLineCX^X
  [4]     -------->  CTetrisFieldLineCX^X
  ...
  [m_nYLen-1] ---->  CTetrisFieldLineCX^X

  ʏł́AindexɂȂ
*/
class CTetrisField
{
protected:
	// ꃉC\NX
	class CTetrisFieldLine
	{
	public:
		CTetrisFieldLine(unsigned int nLen)
		{
			if(nLen > 0){
				m_pBlockState = new ETetrisBlockState[nLen];
			}else{
				m_pBlockState = NULL;
			}
			m_nLen = nLen;
			CTetrisFieldLine::SetBlockState(BLOCKSTATE_NONE);
		}

		virtual ~CTetrisFieldLine()
		{
			delete[] m_pBlockState;
		}

		// TCY`FbN͒PɂsȂ
		ETetrisBlockState GetBlockState(unsigned int nIndex) const
		{
			if(nIndex >= m_nLen){
				return BLOCKSTATE_NONE;
			}else{
				return m_pBlockState[nIndex];
			}
		}

		void SetBlockState(unsigned int nIndex, ETetrisBlockState blockState)
		{
			if(nIndex >= m_nLen){
				return;
			}
			m_pBlockState[nIndex] = blockState;
		}

		void SetBlockState(ETetrisBlockState blockState)
		{
			unsigned int i = 0;
			for(i=0; i<m_nLen; i++){
				m_pBlockState[i] = blockState;
			}
		}

		bool IsFilled() const
		{
			unsigned int i = 0;
			for(i=0; i<m_nLen; i++){
				if(m_pBlockState[i] == BLOCKSTATE_NONE){
					return false;
				}
			}
			return true;
		}

		bool IsEmpty() const
		{
			unsigned int i = 0;
			for(i=0; i<m_nLen; i++){
				if(m_pBlockState[i] != BLOCKSTATE_NONE){
					return false;
				}
			}
			return true;
		}

		unsigned int GetLength() const { return m_nLen; }

	protected:
		unsigned int m_nLen;
		ETetrisBlockState *m_pBlockState;
	};

public:
	typedef unsigned int TetrisFieldResult; // ̒lORŌʂԂ
	static const unsigned int OK            = 0x00;
	static const unsigned int BLOCK_OVERLAP = 0x01;
	static const unsigned int FLOOR_OVERLAP = 0x02;
	static const unsigned int LWALL_OVERLAP = 0x04;
	static const unsigned int RWALL_OVERLAP = 0x08;
	static const unsigned int CEIL_OVERLAP  = 0x10;

	CTetrisField(unsigned int nXLen, unsigned int nYLen, unsigned int nYMarginLen);
	virtual ~CTetrisField();
	void Init();
	TetrisFieldResult GetBlockState(const CPoint2D<int> &rPt, ETetrisBlockState &rBlockState) const;
	TetrisFieldResult SetBlockState(const CPoint2D<int> &rPt, ETetrisBlockState blockState);
	TetrisFieldResult CheckOverlap(const CMinoBlocks &rMinoBlocks) const;
	TetrisFieldResult CheckOverlap(const CPoint2D<int> &rPt) const;
	TetrisFieldResult SetBlockState(const CMinoBlocks &rMinoBlocks);
	bool SetMinoBlocksToStartPosition(CMinoBlocks &rMinoBlocks) const;

	bool IsAllClear() const; // ݑSԂȂtrueԂ
	unsigned int ClearFilledLine(); // CāACԂ
protected:
	void ClearALine(unsigned int nLine);
protected:
	CTetrisFieldLine **m_ppFieldLine;
	unsigned int m_nXLen;
	unsigned int m_nYLen;
	CPoint2D<int> m_ptBase;
private:
	CTetrisField();
	CTetrisField(const CTetrisField&);
	CTetrisField& operator =(const CTetrisField&);
};

#endif // TETRISFIELD_H
