#include "TetrisMgr.h"
#include "SegaRotationMinoSelector.h"
#include "TGMRotationMinoSelector.h"
#include "SRSMinoSelector.h"

ITetrisMgr* GetTetrisMgr()               // TetrisMgrCX^XԂ
{
	return new CTetrisMgr();
}

void ReleaseTetrisMgr(ITetrisMgr* pMgr)  // TetrisMgr
{
	delete pMgr;
}

CTetrisMgr::CTetrisMgr() : m_tetrisField(TETRIS_BLOCKNUM_X, TETRIS_BLOCKNUM_Y, 10),
	m_onFloorChecker(TETRIS_BLOCKNUM_Y, 10)
{
	m_pMinoCurrent = NULL;
	m_pMinoHold = NULL;
	{
		int i = 0;
		for(i=0; i<sizeof(m_aMinoNext)/sizeof(m_aMinoNext[0]); i++){
			m_aMinoNext[i] = NULL;
		}
	}
	m_pMinoSelector = new CSRSMinoSelector();
	m_bGameOver = false;
	m_bHoldDone = false;
	m_nScore = 0;
	m_nLineCleared = 0;
	m_bB2BEnable = false;
	m_tetrisMsg = TETRISMSG_NONE;
	m_rotationSystem = ITetrisMgr::SRS;
	m_bGhostEnable = true;
	m_nInitialLevel = 1;

	// SelectorTetrisField֘At
	m_pMinoSelector->RelateTetrisField(&m_tetrisField);
}

CTetrisMgr::~CTetrisMgr()
{
	delete m_pMinoSelector;
}

ITetrisMgr::EResult CTetrisMgr::OnInit()
{
	m_pMinoCurrent = NULL;
	m_pMinoHold = NULL;
	{
		int i = 0;
		for(i=0; i<sizeof(m_aMinoNext)/sizeof(m_aMinoNext[0]); i++){
			m_aMinoNext[i] = NULL;
		}
	}
	m_tetrisField.Init();
	m_bGameOver = false;
	m_bHoldDone = false;
	m_bMinoOnFloor = false;
	m_bFallTimerReset = false;
	m_nScore = 0;
	m_nLineCleared = 0;
	m_bB2BEnable = false;
	m_tetrisMsg = TETRISMSG_NONE;

	return ITetrisMgr::OK;
}

ITetrisMgr::EResult CTetrisMgr::OnStart()
{
	if(m_bGameOver){
		return ITetrisMgr::ERR_STATE;
	}

	RefreshMinos();

	// MinoX^[gԂɂ
	m_onFloorChecker.ClearCount();
	if((m_pMinoCurrent->Init() & CMino::SUCCESS)
		&& (m_pMinoCurrent->Start() & CMino::SUCCESS)){
		return ITetrisMgr::OK;
	}else{
		return ITetrisMgr::ERR_OTHER;
	}
}

ITetrisMgr::EResult CTetrisMgr::OnLeft()
{
	if(m_bGameOver || !m_pMinoCurrent){
		return ITetrisMgr::ERR_STATE;
	}

	m_pMinoCurrent->Left();
	CheckAndHandleMinoLock();
	return ITetrisMgr::OK;
}

ITetrisMgr::EResult CTetrisMgr::OnRight()
{
	if(m_bGameOver || !m_pMinoCurrent){
		return ITetrisMgr::ERR_STATE;
	}

	m_pMinoCurrent->Right();
	CheckAndHandleMinoLock();
	return ITetrisMgr::OK;
}

ITetrisMgr::EResult CTetrisMgr::OnLRotate()
{
	if(m_bGameOver || !m_pMinoCurrent){
		return ITetrisMgr::ERR_STATE;
	}

	m_pMinoCurrent->LRotate();
	CheckAndHandleMinoLock();
	return ITetrisMgr::OK;
}

ITetrisMgr::EResult CTetrisMgr::OnRRotate()
{
	if(m_bGameOver || !m_pMinoCurrent){
		return ITetrisMgr::ERR_STATE;
	}

	m_pMinoCurrent->RRotate();
	CheckAndHandleMinoLock();
	return ITetrisMgr::OK;
}

ITetrisMgr::EResult CTetrisMgr::OnDown()
{
	if(m_bGameOver || !m_pMinoCurrent){
		return ITetrisMgr::ERR_STATE;
	}

	if(m_pMinoCurrent->Down() & CMino::SUCCESS){
		m_nScore += 1; // ~肽_Z
	}
	CheckAndHandleMinoLock();
	return ITetrisMgr::OK;
}

ITetrisMgr::EResult CTetrisMgr::OnHardDrop()
{
	if(m_bGameOver || !m_pMinoCurrent){
		return ITetrisMgr::ERR_STATE;
	}

	m_pMinoCurrent->HardDrop();
	CheckAndHandleMinoLock();
	return ITetrisMgr::OK;
}

ITetrisMgr::EResult CTetrisMgr::OnHold()
{
	if(m_bGameOver || !m_pMinoCurrent){
		return ITetrisMgr::ERR_STATE;
	}

	if(m_bHoldDone){
		return ITetrisMgr::OK; // xڂHold͕s
	}

	m_bHoldDone = true;
	if(m_pMinoHold){
		// HoldɊMino̓ZbgĂ
		CMino *pMinoTmp = m_pMinoHold;
		m_pMinoHold = m_pMinoCurrent;
		m_pMinoCurrent = pMinoTmp;
	}else{
		// Holdɂ܂Mino̓ZbgĂȂ
		m_pMinoHold = m_pMinoCurrent;
		// Mino̗p
		UpdateMinos();
	}

	m_onFloorChecker.ClearCount();
	if((m_pMinoCurrent->Init() & CMino::SUCCESS)
		&& (m_pMinoCurrent->Start() & CMino::SUCCESS)){
		return ITetrisMgr::OK;
	}else{
		// MinouȂ̂GameOver
		m_bGameOver = true;
		return ITetrisMgr::OK;
	}
}

ITetrisMgr::EResult CTetrisMgr::OnFall()
{
	if(m_bGameOver || !m_pMinoCurrent){
		return ITetrisMgr::ERR_STATE;
	}

	m_pMinoCurrent->Fall();
	CheckAndHandleMinoLock();
	return ITetrisMgr::OK;
}

void CTetrisMgr::GetFallTimerInterval(unsigned int &rTime, bool &rReset)
{
#if 0
	static const unsigned int aTimerInterval[] = {
		1000, 800, 640, 512, 410, 328, 262, 210, 168, 134,
		107, 86, 69, 55, 44, 35, 28, 23, 18, 14, 12, 10
	};
#endif
	static const unsigned int aTimerInterval[] = {
		1000, 800, 640, 512, 410, 328, 262, 210, 168, 134,
		107, 86, 69, 55, 44, 35, 27, 21, 15, 10
	};

	unsigned int nNumLevel = GetLevel();
	if(nNumLevel >= sizeof(aTimerInterval)/sizeof(aTimerInterval[0])){
		rTime = aTimerInterval[sizeof(aTimerInterval)/sizeof(aTimerInterval[0])-1];
	}else{
		rTime = aTimerInterval[nNumLevel-1];
	}

	// MinoubN/ɐڒnĂ
	if(m_bMinoOnFloor){
		if(rTime < 500){
			rTime = 500; // ɋ鎞ɂ500msLock܂ő҂
		}
	}else{
		rReset = false;
	}
	rReset = m_bFallTimerReset;
	m_bFallTimerReset = false;
}

ETetrisBlockState CTetrisMgr::GetBlockState(unsigned int x, unsigned int y)
{
	ETetrisBlockState blockState = BLOCKSTATE_NONE;

	if(m_pMinoCurrent){
		blockState = m_pMinoCurrent->GetBlockState(CPoint2D<int>(x, y));
		
		// Ghost\Ȃꍇ́AGhost͖Ƃɂreturn
		if(!m_bGhostEnable){
			switch(blockState){
			case BLOCKSTATE_CYAN_G:
			case BLOCKSTATE_YELLOW_G:
			case BLOCKSTATE_GREEN_G:
			case BLOCKSTATE_RED_G:
			case BLOCKSTATE_BLUE_G:
			case BLOCKSTATE_ORANGE_G:
			case BLOCKSTATE_MAGENTA_G:
				return BLOCKSTATE_NONE;
			default:
				break;
			}
		}

		if(blockState != BLOCKSTATE_NONE){
			return blockState;
		}
	}

	if(m_tetrisField.GetBlockState(CPoint2D<int>(x, y), blockState) == CTetrisField::OK){
		return blockState;
	}else{
		return BLOCKSTATE_NONE;
	}
}

ITetrisMgr::EResult CTetrisMgr::SetRotationSystem(ERotationSystem system)
{
	if((system != ITetrisMgr::SEGA) && (system != ITetrisMgr::ARS) && (system != ITetrisMgr::SRS)){
		// ςȒlw肳ꂽ
		return ITetrisMgr::ERR_ARG;
	}

	if(system == m_rotationSystem){
		// 邱Ƃ͖
		return ITetrisMgr::OK;
	}

	CMinoSelector *pMinoSelector = NULL;

	switch(system){
	case SEGA:
		pMinoSelector = new CSegaRotationMinoSelector;
		break;
	case ARS:
		pMinoSelector = new CTGMRotationMinoSelector;
		break;
	case SRS:
	default:
		pMinoSelector = new CSRSMinoSelector;
		break;
	}

	if(!pMinoSelector){
		return ITetrisMgr::ERR_OTHER;
	}

	if(m_pMinoSelector){
		delete m_pMinoSelector;
	}

	m_pMinoSelector = pMinoSelector;
	// SelectorTetrisField֘At
	m_pMinoSelector->RelateTetrisField(&m_tetrisField);
	m_rotationSystem = system;

	RefreshMinos();

	return ITetrisMgr::OK;
}

// protected
void CTetrisMgr::CheckAndHandleMinoLock()
{
	if(!m_pMinoCurrent){
		return;
	}

	if(!m_pMinoCurrent->IsLocked()){
		// ܂ɗĂȂ
		CPoint2D<int> pt;
		if(m_pMinoCurrent->IsOnFloor(pt)){
			// ptAݐڂĂubNW
			m_bMinoOnFloor = true;
			if(m_onFloorChecker.AddCountAndCheckTimerReset(pt.y)){
				// ^C}ZbgKv
				m_bFallTimerReset = true;
			}
		}else{
			m_bMinoOnFloor = false;
			m_onFloorChecker.NotOnFloor();
		}
		return; 
	}

	unsigned int nFilledLine = m_tetrisField.ClearFilledLine();

	m_tetrisMsg = TETRISMSG_NONE;

	{
		// _vZ
		unsigned int nScore = 0;
		if(m_pMinoCurrent->IsTSpin()){
			if(m_bB2BEnable){
				switch(nFilledLine){
				case 0:
				default:
					if(m_pMinoCurrent->IsWallKick()){
						// TSpin(ᓾ_)
						nScore += 100;
					}else{
						// TSpin(_)
						nScore += 400;
					}
					m_tetrisMsg = TETRISMSG_TSPIN;
					break;
				case 1:
					if(m_pMinoCurrent->IsWallKick()){
						// TSpinSingle(ᓾ_) + B2B
						nScore += 300;
					}else{
						// TSpinSingle(_) + B2B
						nScore += 1200;
					}
					m_tetrisMsg = TETRISMSG_TSPINSINGLE_B2B;
					m_bB2BEnable = true;
					break;
				case 2: // TSpinDouble + B2B
					nScore += 1800; m_tetrisMsg = TETRISMSG_TSPINDOUBLE_B2B; m_bB2BEnable = true; break;
				case 3: // TSpinTriple + B2B
					nScore += 2400; m_tetrisMsg = TETRISMSG_TSPINTRIPLE_B2B; m_bB2BEnable = true; break;
				}
			}else{
				switch(nFilledLine){
				case 0:
				default:
					if(m_pMinoCurrent->IsWallKick()){
						// TSpin(ᓾ_)
						nScore += 100;
					}else{
						// TSpin(_)
						nScore += 400;
					}
					m_tetrisMsg = TETRISMSG_TSPIN;
					break;
				case 1:
					if(m_pMinoCurrent->IsWallKick()){
						// TSpinSingle(ᓾ_)
						nScore += 200;
					}else{
						// TSpinSingle(_)
						nScore += 800;
					}
					m_tetrisMsg = TETRISMSG_TSPINSINGLE;
					m_bB2BEnable = true;
					break;
				case 2: // TSpinDouble
					nScore += 1200; m_tetrisMsg = TETRISMSG_TSPINDOUBLE; m_bB2BEnable = true; break;
				case 3: // TSpinTriple
					nScore += 1600; m_tetrisMsg = TETRISMSG_TSPINTRIPLE; m_bB2BEnable = true; break;
				}
			}
		}else{
			switch(nFilledLine){
			case 0:
			default:
				break;
			case 1: // Single
				nScore += 100; m_tetrisMsg = TETRISMSG_SINGLE; m_bB2BEnable = false; break;
			case 2: // Double
				nScore += 300; m_tetrisMsg = TETRISMSG_DOUBLE; m_bB2BEnable = false; break;
			case 3: // Triple
				nScore += 500; m_tetrisMsg = TETRISMSG_TRIPLE; m_bB2BEnable = false; break;
			case 4: // Tetris
				if(m_bB2BEnable){
					// Back To Back Tetris
					nScore += 1200;
					m_tetrisMsg = TETRISMSG_TETRIS_B2B;
				}else{
					nScore += 800;
					m_tetrisMsg = TETRISMSG_TETRIS;
				}
				m_bB2BEnable = true;
				break;
			}
		}

		nScore *= GetLevel();

		if(m_tetrisField.IsAllClear()){
			nScore *= 10;
			m_tetrisMsg = TETRISMSG_ALLCLEAR;
		}

		m_nScore += nScore;
	}

	// HardDropZ(*2_)
	m_nScore += m_pMinoCurrent->GetHardDropHeight() * 2;

	m_nLineCleared += nFilledLine;
	m_bHoldDone = false; // xMinoŒ肳ꂽ̂ŁAHold

	// Mino̗p
	UpdateMinos();

	m_onFloorChecker.ClearCount();
	if((m_pMinoCurrent->Init() & CMino::SUCCESS)
		&& (m_pMinoCurrent->Start() & CMino::SUCCESS)){
		return;
	}else{
		// MinouȂ̂GameOver
		m_bGameOver = true;
		return;
	}
}

void CTetrisMgr::RefreshMinos()
{
	if(!m_pMinoSelector){
		return;
	}

	// Mino܂I
	m_pMinoCurrent = m_pMinoSelector->SelectMino();
	m_pMinoHold = NULL; // Hold͋ɂ
	{
		int i = 0;
		for(i=0; i<sizeof(m_aMinoNext)/sizeof(m_aMinoNext[0]); i++){
			m_aMinoNext[i] = m_pMinoSelector->SelectMino();
		}
	}
}

void CTetrisMgr::UpdateMinos()
{
	if(!m_pMinoSelector){
		return;
	}

	m_pMinoCurrent = m_aMinoNext[0];
	{
		int i = 0;
		for(i=0; i<sizeof(m_aMinoNext)/sizeof(m_aMinoNext[0])-1; i++){
			m_aMinoNext[i] = m_aMinoNext[i+1];
		}
		m_aMinoNext[i] = m_pMinoSelector->SelectMino();
	}
}

